package org.szte.code.client;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.swing.JPanel;
import javax.swing.JTree;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeCellRenderer;

import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonStreamParser;

/**
 *  2013, SZTE, CIRENE project, GPL licence http://www.gnu.org/licenses/gpl-2.0.html
 *  
 *  @author Rcz, Rbert
 *
 */


public class CoveragePanel extends JPanel {

	private static final long serialVersionUID = 9216559768517717032L;
	
	protected class MethodInfo {
		private int id;
		private String header;
		private int coverage;
		
		public MethodInfo(int id, String header) {
			this.id = id;
			this.header = header;
		}
		
		public void setCoverage(int coverage) {
			this.coverage = coverage;
		}
		
		public int getId() {
			return id;
		}

		public String getHeader() {
			return header;
		}

		public int getCoverage() {
			return coverage;
		}
		
		public String toString() {
			return "(covered: " + getCoverage() + ", id: " + getId() + ") " + getHeader();
		}
	}

	protected class ClassInfo extends ArrayList<MethodInfo> {
		private class NameComparator implements Comparator<MethodInfo> {
			@Override
			public int compare(MethodInfo arg0, MethodInfo arg1) {
				return arg0.getHeader().compareTo(arg1.getHeader());
			}
		}
		
		private static final long serialVersionUID = 7570066314831054168L;
		private String name;
		
		public ClassInfo(String name) {
			this.name = name;
		}
		
		public void sort() {
			java.util.Collections.sort(this, new NameComparator());
		}
		
		public String getName() {
			return name;
		}
		
		public int getCoveredMethodNum() {
			int covered = 0;
			for (MethodInfo method : this) {
				if (method.getCoverage() > 0) {
					covered++;
				}
			}
			return covered;
		}
		
		public String toString() {
			return name + " (covered: " + getCoveredMethodNum()*100.0f/size() + "%)";
		}
	}

	protected class ApplicationInfo {
		private Map<Integer, MethodInfo> methods = new HashMap<Integer, MethodInfo>();
		private Map<String, ClassInfo> classes = new HashMap<String, ClassInfo>();
		private String name;
		private int coveredMethod = 0;
		
		public ApplicationInfo(String name) {
			this.name = name;
		}
		
		public void addMethod(MethodInfo method, String className) {
			if (!classes.containsKey(className)) {
				classes.put(className, new ClassInfo(className));
			}
			classes.get(className).add(method);
			methods.put(method.getId(), method);
			if (method.getCoverage() > 0) {
				coveredMethod++;
			}
		}
		
		public void updateCoverage(int methodId, int coverage) {
			if (methods.containsKey(methodId)) {
				MethodInfo method = methods.get(methodId);
				if (method.getCoverage() != coverage) {
					if (coverage == 0) {
						coveredMethod--;
					} else if (method.getCoverage() == 0) {
						coveredMethod++;
					}
				}
				method.setCoverage(coverage);
			}
		}
		
		public String getName() {
			return name;
		}
		
		public Map<String, ClassInfo> getCoverage() {
			return classes;
		}

		public int getCoveredMethodNum() {
			return coveredMethod;
		}
		
		public String toString() {
			return name + " (covered: " + getCoveredMethodNum()*100.0f/methods.size() + "%)";
		}
	}
	
	class TreeRenderer extends DefaultTreeCellRenderer {
		private static final long serialVersionUID = -1801898411633675595L;
		private final Color coveredColor = new Color(10, 110, 10);
		private final Color notCoveredColor = Color.gray;
		private final Font coveredFont = new Font(Font.SERIF, Font.BOLD, 13);
		private final Font notCoveredFont = new Font(Font.SERIF, Font.PLAIN, 12);
		
		
		public Component getTreeCellRendererComponent(JTree tree, Object value, boolean sel, boolean expanded, boolean leaf, int row, boolean hasFocus) {
			Object userObject = ((DefaultMutableTreeNode)value).getUserObject();
			if (userObject instanceof MethodInfo) {
				if (((MethodInfo)userObject).getCoverage() > 0) {
					setTextNonSelectionColor(coveredColor);
					setFont(coveredFont);
				} else {
					setTextNonSelectionColor(notCoveredColor);
					setFont(notCoveredFont);
				}
			} else if (userObject instanceof ClassInfo) {
				if (((ClassInfo)userObject).getCoveredMethodNum() > 0) {
					setTextNonSelectionColor(coveredColor);
					setFont(coveredFont);
				} else {
					setTextNonSelectionColor(notCoveredColor);
					setFont(notCoveredFont);
				}
			} else if (userObject instanceof ApplicationInfo) {
				if (((ApplicationInfo)userObject).getCoveredMethodNum() > 0) {
					setTextNonSelectionColor(coveredColor);
					setFont(coveredFont);
				} else {
					setTextNonSelectionColor(notCoveredColor);
					setFont(notCoveredFont);
				}
			} else {
				setTextNonSelectionColor(Color.black);
				setFont(new Font(Font.SERIF, Font.PLAIN, 12));
			}

			super.getTreeCellRendererComponent(tree, value, sel, expanded, leaf, row, hasFocus);
			return this;
		}
	}
	
	private Map<String, ApplicationInfo > applications = new HashMap< String, ApplicationInfo >();
	private String currentTimestamp = "";
	
	CoveragePanel() {
		setLayout(new BorderLayout());
	}
	
	public void parseApplicationMethodCsv(String fileName) throws IOException {
		BufferedReader br = new BufferedReader( new FileReader(fileName));
		String strLine = "";
		if ((strLine = br.readLine()) != null) {
			String[] columns = strLine.split(";");
			if (columns.length == 2 && "ProjectName".compareTo(columns[0]) == 0) {
				ApplicationInfo ainfo = new ApplicationInfo(columns[1]);
				while( (strLine = br.readLine()) != null) {
					String[] methodInfo = strLine.split(";");
					if (methodInfo.length == 3) {
						int id = Integer.parseInt(methodInfo[0]);
						MethodInfo method = new MethodInfo(id, methodInfo[1]);
						ainfo.addMethod(method, methodInfo[2]);
					}
				}
				
				if (applications.containsKey(ainfo.getName())) {
					ApplicationInfo eainfo = applications.get(ainfo.getName());
					for (Entry<String, ClassInfo> classInfo : eainfo.getCoverage().entrySet()) {
						for (MethodInfo method : classInfo.getValue()) {
							ainfo.updateCoverage(method.getId(), method.getCoverage());
						}
					}
				}
				
				applications.put(ainfo.getName(), ainfo);
				
				updateTree();
			}
		}
		
		br.close();
	}
	
	public void updateCoverageInformation(String coverage) {
		JsonStreamParser parser = new JsonStreamParser(coverage);
		if (parser.hasNext()) {
			JsonObject message = parser.next().getAsJsonObject();
			if (message.has("timestamp")) {
				currentTimestamp = message.get("timestamp").getAsString();
			}

			if (message.has("data")) {
				JsonObject data = message.getAsJsonObject("data");
				for (Entry<String , JsonElement> application : data.entrySet()) {
					boolean exist = true;
					ApplicationInfo app = applications.get(application.getKey());
					if (app == null) {
						app = new ApplicationInfo(application.getKey());
						exist = false;
					}
					
					JsonArray coverageList = application.getValue().getAsJsonArray();
					int methodNum = coverageList.size();
					for (int i=0; i<methodNum; ++i) {
						if (exist) {
							app.updateCoverage(i, coverageList.get(i).getAsInt());
						} else {
							MethodInfo method = new MethodInfo(i, "Method-" + i);
							method.setCoverage(coverageList.get(i).getAsInt());
							app.addMethod(method, "undefined");
						}
					}
					if (!exist) {
						applications.put(app.getName(), app);
					}
				}
			}
			
			updateTree();
		}
	}

	private void updateTree() {
		removeAll();

		DefaultMutableTreeNode top = new DefaultMutableTreeNode("Coverage information - " + currentTimestamp);
		List<String> sorted = new ArrayList<String>(applications.keySet());
		java.util.Collections.sort(sorted);
		
		for (String applicationName : sorted) {
			DefaultMutableTreeNode applicationNode = new DefaultMutableTreeNode(applications.get(applicationName));
		    top.add(applicationNode);
		    Map<String, ClassInfo> classMap = applications.get(applicationName).getCoverage();
		    List<String> sortedClassName = new ArrayList<String>(classMap.keySet());
			java.util.Collections.sort(sortedClassName);
			for (String classInfoName : sortedClassName) {
				DefaultMutableTreeNode classNode = new DefaultMutableTreeNode(classMap.get(classInfoName));
				applicationNode.add(classNode);
				ClassInfo classInfo = classMap.get(classInfoName);
				classInfo.sort();
				for (MethodInfo method : classInfo) {
					classNode.add(new DefaultMutableTreeNode(method));
				}
			}
		}
		JTree tree = new JTree(top);
		tree.setCellRenderer(new TreeRenderer());
		add(tree, BorderLayout.CENTER);
		
		updateUI();
	}
}
